/* network.c   linux-c
 * MG server
 *
 */

#include "commands.h"
#include "network.h"
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

int inet_listener;
int unix_listener;
int mg_sync = 0;

void init_packet( MG_packet *packet, uint8 type )
{
	packet->private = packet->size = 0;
	packet->type = type;
}

// private is used as a read index
void extract_bytes( void *dest, MG_packet *packet, int count )
{
	if( count > packet->size  )
		count = packet->size ;
	memcpy( dest, &packet->data[packet->private], count );
	packet->private += count;  
}

void dump_packet( MG_packet *packet )
{
	int i;
	for( i = 0; i < packet->size + 3; i++ )
		printf("%x ", *(((unsigned char*)packet) + i));
	printf("\n");
}

void store_bytes( MG_packet *packet, const void *buf, int count )
{
	memcpy( &packet->data[packet->size], buf, count );
	packet->size += count;
}


uint8 extract_uint8( MG_packet *packet )
{
	return packet->data[packet->private++];
}

void store_uint8( MG_packet *packet, uint8 i )
{
	packet->data[packet->size++] = i;
}

int extract_uint16( MG_packet *packet )
{
	int l, h;
	h = (int)packet->data[packet->private++] << 8;
	l = (int)packet->data[packet->private++];
	return h + l;
}

void store_uint16( MG_packet *packet, int i )
{
	packet->data[packet->size++] = ( i >> 8 ) & 0xff;
	packet->data[packet->size++] = i & 0xff;
}


int extract_sint16(MG_packet *packet)
{
	int i, l, h;
	h = ( (int)packet->data[packet->private++] & 0x7f ) << 8;
	l = (int)packet->data[packet->private++];
	i = h + l;
	if( packet->data[packet->private - 2] & 0x80 )
		i = -i;
	return i;
}

void store_sint16( MG_packet *packet, int i )
{
	int j = abs( i );
	packet->data[packet->size++] = ( j >> 8 ) & 0x7f;
	packet->data[packet->size++] = j & 0xff;
	if( i < 0 )
		packet->data[packet->size - 2] |= 0x80;
}

unsigned int extract_uint32( MG_packet *packet )
{
	unsigned int x1, x2, x3, x4;
	x1 = packet->data[packet->private++] << 24;
	x2 = packet->data[packet->private++] << 16;
	x3 = packet->data[packet->private++] << 8;
	x4 = packet->data[packet->private++];
	return x1 + x2 + x3 + x4;
}

void store_uint32( MG_packet *packet, unsigned int i)
{
	packet->data[packet->size++] = (i >> 24) & 0xff;
	packet->data[packet->size++] = (i >> 16) & 0xff;
	packet->data[packet->size++] = (i >> 8) & 0xff;
	packet->data[packet->size++] = i & 0xff;
}

int extract_sint32(MG_packet *packet)
{
	int i, x1, x2, x3, x4;
	x1 = (packet->data[packet->private++] & 0x7f) << 24;
	x2 = packet->data[packet->private++] << 16;
	x3 = packet->data[packet->private++] << 8;
	x4 = packet->data[packet->private++];
	i = x1+x2+x3+x4;
	if( packet->data[packet->private - 4] & 0x80 )
		i = -i;
	return i;
}

void store_sint32(MG_packet *packet, int i)
{
	int j = abs(i);
	packet->data[packet->size++] =  j >> 24;
	packet->data[packet->size++] = j >> 16;
	packet->data[packet->size++] = j >> 8;
	packet->data[packet->size++] = j;
	if( i < 0 )
		packet->data[packet->size - 4] |= 0x80;
}


// 4 bytes
ggi_coord extract_ggicoord( MG_packet *packet )
{
	ggi_coord coord;
	coord.x = extract_sint16( packet );
	coord.y = extract_sint16( packet );
	return coord;
}

// 4 bytes
void store_ggicoord( MG_packet *packet, ggi_coord coord )
{
	store_sint16( packet, coord.x );
	store_sint16( packet, coord.y );
}


// 24 bytes
void extract_ggimode( ggi_mode *mode, MG_packet *packet )
{
	mode->frames = extract_uint32( packet );   
	mode->visible = extract_ggicoord( packet );
	mode->virt = extract_ggicoord( packet );    
	mode->size = extract_ggicoord( packet );   
	mode->graphtype = extract_uint32( packet );
	mode->dpp = extract_ggicoord( packet );    
}


// 24 bytes
void store_ggimode( MG_packet *packet, ggi_mode *mode )
{
	store_uint32( packet, mode->frames );
	store_ggicoord( packet, mode->visible );
	store_ggicoord( packet, mode->virt );
	store_ggicoord( packet, mode->size );
	store_uint32( packet, mode->graphtype );
	store_ggicoord( packet, mode->dpp );
}

// 8 bytes
ggi_color extract_ggicolor( MG_packet *packet )
{
	ggi_color col;
	col.r = extract_uint16( packet );
	col.g = extract_uint16( packet );
	col.b = extract_uint16( packet );
	col.a = extract_uint16( packet );
	return col;
}

// 8 bytes
void store_ggicolor( MG_packet *packet, ggi_color *col )
{
	store_uint16( packet, col->r );
	store_uint16( packet, col->g );
	store_uint16( packet , col->b );
	store_uint16( packet, col->a );
}


int extract_giievent( gii_event *event, MG_packet *packet)
{
	event->any.type = extract_uint8( packet );
	event->any.size = extract_uint8( packet );
	event->any.origin = extract_uint32( packet );
	event->any.target = extract_uint32( packet );
	event->any.time.tv_sec = extract_uint32( packet );
	event->any.time.tv_usec = extract_uint32( packet );
	switch( event->any.type ){
	case evCommand:
	case evInformation:
	{
		int i;
		event->cmd.code = extract_uint32( packet );
		i = event->any.size - packet->private;
		extract_bytes( event->cmd.data, packet, i );
		break;
	}
	case evExpose:
		event->expose.x = extract_uint32( packet );
		event->expose.y = extract_uint32( packet );
		event->expose.h = extract_uint32( packet );
		event->expose.w = extract_uint32( packet );
		break;
	case evValRelative:
	case evValAbsolute:
	{
		int i;
		event->val.first = extract_uint32( packet );
		event->val.count = extract_uint32( packet );
		for( i = 0 ; i < 32; i++)
			event->val.value[i] = extract_sint32( packet );
	}
	case evKeyPress:
	case evKeyRelease:
	case evKeyRepeat:
		// MODIFIER_DATA
		event->key.modifiers = extract_uint32( packet );
		// normal data
		event->key.sym = extract_uint32( packet );
		event->key.label = extract_uint32( packet );
		event->key.button = extract_uint32( packet );
		break;
	case evPtrRelative:
	case evPtrAbsolute:
		// normal data
		event->pmove.x = extract_uint32( packet );
		event->pmove.y = extract_uint32( packet );
		event->pmove.z = extract_uint32( packet );
		event->pmove.wheel = extract_uint32( packet );
		break;
	case evPtrButtonPress:
	case evPtrButtonRelease:
		// normal data
		event->pbutton.button = extract_uint32( packet );
		break;
	default:
		break;
	}
	return 0;
}

int store_giievent( MG_packet *packet, gii_event *event)
{
	store_uint8( packet, event->any.type );
	store_uint8( packet, event->any.size );
	store_uint32( packet, event->any.origin );
	store_uint32( packet, event->any.target );
	store_uint32( packet, event->any.time.tv_sec );
	store_uint32( packet, event->any.time.tv_usec );
	switch( event->any.type ){
	case evCommand:
	case evInformation:
	{
		int i;
		store_uint32( packet, event->cmd.code );
		i  = event->any.size - packet->size;
		store_bytes( packet, event->cmd.data, i );
		break;
	}
	case evExpose:
		store_uint32( packet, event->expose.x );
		store_uint32( packet, event->expose.y );
		store_uint32( packet, event->expose.h );
		store_uint32( packet, event->expose.w );
		break;
	case evValRelative:
	case evValAbsolute:
	{
		int i;
		store_uint32( packet, event->val.first );
		store_uint32( packet, event->val.count );
		for( i = 0 ; i < 32; i++)
			store_sint32( packet, event->val.value[i] );
	}
	case evKeyPress:
	case evKeyRelease:
	case evKeyRepeat:
		// MODIFIER_DATA
		store_uint32( packet, event->key.modifiers );
		// normal data
		store_uint32( packet, event->key.sym );
		store_uint32( packet, event->key.label );
		store_uint32( packet, event->key.button );
		break;
	case evPtrRelative:
	case evPtrAbsolute:
		// normal data
		store_uint32( packet, event->pmove.x );
		store_uint32( packet, event->pmove.y );
		store_uint32( packet, event->pmove.z );
		store_uint32( packet, event->pmove.wheel );
		break;
	case evPtrButtonPress:
	case evPtrButtonRelease:
		// normal data
		store_uint32( packet, event->pbutton.button );
		break;
	default:
		break;
	}
	return 0;
}



int init_inet( void )
{  
	int s = 0;
	struct sockaddr_in servaddr;
	
	if( ( s = socket(AF_INET, SOCK_STREAM,0)) < 0 ){
		perror("inet:could not create socket\n");
		return -1;
	}
	
	bzero( &servaddr, sizeof(servaddr) );
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(MG_PORT);
	servaddr.sin_addr.s_addr = INADDR_ANY; 
//       	bzero( &(servaddr.sin_zero), 8 );

	if( bind(s, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0){
		perror("bind failed\n");
		return -1;
	}
	
	if( listen( s, BACKLOG ) == -1){
		perror("listen failed\n");
		return -1;
	}
	printf("listening on socket %d\n", s);
	return s;
}

int init_unix(void)
{
	int s;
	struct sockaddr_un servaddr;
	
	if( (s = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0){
		perror("unix:could not create socket\n");
		return -1;
	}
	
	unlink("/tmp/mg_foo");
	bzero( &servaddr, sizeof( servaddr ));
	servaddr.sun_family = AF_LOCAL;
	strcpy(servaddr.sun_path, "/tmp/mg_foo");

	if( bind(s, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0){
		perror("unix: bind failed");
		return -1;
	}
	if( listen( s, BACKLOG ) == -1){
		perror("listen failed\n");
		return -1;
	}
	printf("listening on socket %d\n", s);
	return s;
}


void init_network(void)
{
	inet_listener = init_inet();
	unix_listener = init_unix();
}

int add_listeners( fd_set *set )
{
	FD_SET(unix_listener, set);
	FD_SET(inet_listener, set);
	printf(" %d and %d added\n", unix_listener, inet_listener );
	return MAX(unix_listener, inet_listener);
}

int sendcounter = 0;

int sent_bytes = 0;

int queue_packet( communication *c, MG_packet *packet )
{
	if( c->out.type != 0 ){
		printf("error queueing packet of type %d", packet->type);
		printf(", buffer contains %d\n",  c->out.type);
		return -1;
	}
	memcpy( &c->out, packet, packet->size + 3 );
	return 0;
}

int send_packet( int socket, MG_packet *packet )
{
	// {todo} convert header of packet before sending
	int err;
	int count = packet->size + 3;

//	printf("network.c: send_packet: ");
//	dump_packet( packet );
	
	err = send( socket, (void *)packet, count, 0 );
	if( err == 0 ){
		printf("network.c: connection lost\n");
		return -1;
	}
	if( err == -1 ){
		printf("network.c: send error  %d\n", errno);
		return -2;
	}
	if( err != count ){
		printf("network.c: data skipped\n");
		return -3;
	}
	sent_bytes += err;
	//printf("network.c: sent %d bytes\n", err );
	return 0;
}


int write_packet( MG_connection *conn, MG_packet *packet )
{
	int err;
	pthread_mutex_lock(&(conn->send_mutex));
	err = send_packet(conn->socket, packet);
	pthread_mutex_unlock(&(conn->send_mutex));
	return err;
}

int received_bytes = 0;

int recv_bytes( int s, void *buf, int len )
{
	int err;
	if( len == 0 )
		return 0;
	err = recv( s, buf, len, 0);
	if( err == 0 ){
		printf("network.c: connection broken\n");
		return -1;
	}
	if( err == -1 ){
		perror("network.c: recv error\n");
		return -2;
	}
	if( err != len ){
		printf("network.c: data lost [ %d ], received %d of %d bytes\n", errno,
		       err, len);
		return -3;
	}
	received_bytes += err;
	return 0;
}


int load_bytes( int s, void *buf, int len )
{
	int err;
	if( len == 0 )
		return 0;
	err = recv( s, buf, len, 0);
	if( err == 0 ){
		printf("network.c: connection broken\n");
		return -1;
	}
	if( err == -1 ){
		perror("network.c: recv error\n");
		return -2;
	}
	return err;
}

int flush_buffer( communication *c )
{
	int err;
	if( c->out.type == 0 )
		return 0;
	err = send( c->socket,
		    (void*)((uint8*)&c->out + c->sent),
		    c->out.size + 3 - c->sent,
		    0);
	if( err == 0){
		printf("network.c: connection lost\n");
		return -1;
	}
	if( err == -1 ){
		printf("network.c: send error  %d\n", errno);
		return -2;
	}
	c->sent += err;
	if( c->sent == c->out.size + 3 ){
		c->out.type = 0;
		c->sent = 0;
	}
	return 0;	
}


int load_packet( communication *c )
{
	int err;
	if( c->read < 3 ){
		err = load_bytes(c->socket, 
				 (uint8*)&c->packet + c->read,
				 3 - c->read);
		if( err < 0 ){
//			printf("load_header -> %d\n");
			return err;
		}
		c->read += err;
	}
	if( (c->read >= 3) && ( c->read < c->packet.size + 3 ) ){
		err = load_bytes(c->socket, 
				 (uint8*)&c->packet + c->read,
				 c->packet.size + 3 - c->read);
		if( err < 0 ){
//			printf("load_body -> %d\n");
			return err;
		}
		c->read += err;
	}
	if( (c->read >= 3) && (c->read == c->packet.size + 3)){
//		printf("got a whole packet \n");
		return 0; 
	}
//	printf("data is still missing\n");
	return 1;
}

void clear_packet( communication *c )
{
	c->read = 0;
	c->packet.type = 0;

}


int recvcounter = 0;


int recv_packet( int socket, MG_packet *packet )
{
	MG_packet sp;
	int err;
	uint8 header[3];
//	printf("network.c: recv_packet no %d\n", recvcounter++);
	err = recv_bytes( socket, header, 3 );
	if( err )
		return err;
	packet->private = header[0];
	packet->size = header[1];
	packet->type = header[2];
//	printf("network.c: header: %x %x %x\n", header[0], header[1], header[2]);
	err = recv_bytes( socket, packet->data, packet->size );
	//printf("network.c: received %d bytes\n", packet->size );
	if( mg_sync ){
		init_packet( &sp, 33 );
		send_packet( socket, &sp );
	}
//	dump_packet(packet);
	return err;
}


// store is a bad name - conflicts with the above!
MG_packet *store_packet( MG_connection *conn, MG_packet* packet )
{
	int i = conn->lastpos + 1;
	if( i >= ELEMENTS )
		i = 0;
//	printf("storing at position %d\n", i );
	memcpy( &conn->buffer[i], packet, packet->size );
	conn->lastpos = i;
	return &conn->buffer[i];
}

// always check the oldest packets first!
MG_packet *scan_buffers( int type, MG_connection *conn )
{
	int i = conn->lastpos + 1;
	while(-1){
		if( i >= ELEMENTS )
			i = 0;
		if( i == conn->lastpos )
			return NULL;
		if( conn->buffer[i].type == type )
			return  &conn->buffer[i]; 
		i++;
	}
}


int time_left(struct timeval *timeout)
{
	if( timeout == NULL )
		return 1;
	return (timeout->tv_sec || timeout->tv_usec);
}

void trim_timeout(struct timeval *timeout, struct timeval *starttime )
{
	long usec;
        struct timeval t;
	if( timeout == NULL )
		return;
        gettimeofday( &t, NULL );
        usec = (t.tv_sec - starttime->tv_sec)*1000000 + 
                t.tv_usec - starttime->tv_usec;
        usec = timeout->tv_sec * 1000000 + timeout->tv_usec - usec;
	if( usec < 0 )
		usec = 0;
        timeout->tv_sec = usec / 1000000;
        timeout->tv_usec = usec % 1000000;
}


int MGtimedwait( pthread_cond_t *cond,
		 pthread_mutex_t *mutex,
		 struct timeval *timeout )
{
	int n;
	struct timespec  time;
	if( timeout == NULL ){
		pthread_mutex_lock( mutex );
		pthread_cond_wait( cond, mutex );
		pthread_mutex_unlock( mutex );
		return 0;
	} else {
		time.tv_sec = timeout->tv_sec;
		time.tv_nsec = timeout->tv_usec * 1000;
		n = pthread_cond_timedwait( cond, mutex, &time );
		if( n == ETIMEDOUT )
			return -1;
		return 0;
	}
}

void MGbroadcast( pthread_cond_t *cond,  pthread_mutex_t *mutex )
{
	pthread_mutex_lock( mutex );
	pthread_cond_broadcast( cond );
	pthread_mutex_unlock( mutex );
}

//return type of packet
//if packet is of type "type" store in packet otherwise in buffer
int scan_incoming(int type, MG_connection *conn, 
		  struct timeval *timeout, MG_packet *packet )
{
	int err;
	fd_set sockset;
	FD_ZERO( &sockset );
	FD_SET( conn->socket, &sockset );
	err = select( conn->socket + 1, &sockset, NULL, NULL, timeout );
	if(err == 0){
//		printf("select timeout\n");
		return 0;
	}
	if(err < 0 ){
		printf("select error\n");
		return err;
	}
	err = recv_packet( conn->socket, packet );
	if( err )
		return err;
	if( packet->type != type )
		store_packet( conn, packet );
	return packet->type;
}

// return -1 on timeout
int wait_packet( int type, MG_connection *conn, 
		 MG_packet *packet, struct timeval *timeout )
{
	MG_packet *result;
	int received;
	struct timeval starttime;
	gettimeofday( &starttime, NULL );
	do{
		pthread_mutex_lock( &conn->buffer_mutex );
		if( (result = scan_buffers(type, conn)) != NULL ){
			memcpy( packet, result, result->size );
			result->type = 0;
			pthread_mutex_unlock( &conn->buffer_mutex );
			return 0;
		}
		if( conn->recv_flag ){
			int n;
			printf("waiting ...\n");
			pthread_mutex_unlock( &conn->buffer_mutex );
			n = MGtimedwait( &conn->recv_status,
					 &conn->wait_mutex,
					 timeout);
			if( n == -1 )
				return -1;
			printf("wait done\n");
			trim_timeout( timeout, &starttime );
			continue;
		}
		conn->recv_flag = 1;
		pthread_mutex_unlock( &conn->buffer_mutex );
		received = scan_incoming(type, conn, timeout, packet );
		MGbroadcast( &conn->recv_status, &conn->wait_mutex );
		conn->recv_flag = 0;
		if(received == 0 )
			return -1;
		if( received == type )
			return 0;
		trim_timeout( timeout, &starttime );
	}while( time_left(timeout));
	printf("timeout!\n");
	return -1;
}


int get_packet(int type, MG_connection *conn, MG_packet *packet)
{
//	printf("get_packet started\n");
	return wait_packet( type, conn, packet, NULL );
//	printf("get_packet stopped\n");
}

//{todo - make accept signal-proof}
int accept_connection( int s )
{
	struct sockaddr new_addr;
	unsigned int size;
	int client_fd = accept( s, &new_addr, &size );
	fcntl(client_fd, F_SETFL, O_NONBLOCK );
	if( client_fd == -1 ){
		printf(" Could not accept new connection\n");
	}
	return client_fd;
}


int new_connection( fd_set *set )
{
	if( FD_ISSET( inet_listener, set ) ){
		FD_CLR( inet_listener, set );
		return inet_listener;
	}
	if( FD_ISSET( unix_listener, set) ){
		FD_CLR( unix_listener, set);
		return unix_listener;
	}
	return 0;
}
