/* regions.c
 * MG server
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include "regions.h"


int in_region( int x, int y, region *r )
{
	return ( x >= r->x &&
		 x < r->x + r->w &&
		 y >= r->y &&
		 y < r->y + r->h );
}

int covers( region *a, region *b )
{
	int r;
	r = !( (a->x >= b->x + b->w ) ||
	       (b->x >= a->x + a->w ) ||
	       (a->y >= b->y + b->h ) ||
	       (b->y >= a->y + a->h ) );
	return r;
}

// this gives a list: left->right
region *split_vertical( region *source, int x )
{
	region *result = NULL;
	if( x <= source->x || x >= source->x + source->w ){
		printf("no vertical split\n");
		return NULL;
	}
	add_region( &result, create_region( x,
					    source->y,
					    source->x + source->w - x,
					    source->h ) );
	add_region( &result, create_region( source->x,
					    source->y,
					    x - source->x,
					    source->h ) );
	return result;
	
}

// this gives a list top->bottom
region *split_horisontal( region *source, int y )
{
	region *result = NULL;
	if( y <=source->y || y >= source->y + source->h ){
		printf("no horizontal split\n");
		return NULL;
	}
	add_region( &result, create_region( source->x,
					    y,
					    source->w,
					    source->y + source->h - y ));
	add_region( &result, create_region( source->x,
					    source->y,
					    source->w,
					    y - source->y ) );
	return result;		    
}

// this should return (a)-(b)
region *difference_single( region *list, region *mask )
{
	region *diff = NULL;

	if( mask->w == 0 || mask->h == 0 )
		return duplicate( list );

	while( list != NULL ){
		region *split;
		region *current = create_region( list->x, list->y, 
						 list->w, list->h );

		if( !covers( mask, current ) ){
			grow( &diff, current );
			list = list->next;
			continue;
		}

		// cut along leftmost edge 
		if( mask->x > current->x ){
			split = split_vertical( current, mask->x );
			grow( &diff, remove_region( &split, split ) );
			destroy_region( current );
			current = split;
		}

		// cut along rightmost edge
		if( (current->x + current->w) > (mask->x + mask->w)){
			split = split_vertical( current, mask->x + mask->w);
			destroy_region( current );
			current = remove_region( &split, split );
			grow( &diff, split );
		}

		// cut along topmost edge
		if( mask->y > current->y ){
			split = split_horisontal( current, mask->y );
			add_region( &diff, remove_region( &split, split ) );
			destroy_region( current );
			current = split;
		}

		// cut along bottom edge
		if( current->y + current->h > mask->y + mask->h){
			split = split_horisontal( current, mask->y + mask->h );
			destroy_region( current );
			current = remove_region( &split, split );
			grow( &diff, split );
		}
		destroy_region( current );
		list = list->next;
	}
	return diff;
}

region *difference( region *list, region *masks )
{
	region *diff = duplicate(list);
	while(masks != NULL ){
		region *new_diff = difference_single(diff, masks );
		destroy_regions( &diff );
		diff = new_diff;
		masks = masks->next;
	}
	return diff;
}       


int joinable( region *a, region *b )
{
	if( a == NULL || b == NULL )
		return 0;
	return  ( ( a->x == b->x ) && 
		  ( a->w == b->w ) &&
		  ( ( a->y + a->h == b->y ) || ( b->y + b->h == a->y ) ) ) ||
		( ( a->y == b->y ) && 
		  ( a->h == b->h ) &&
		  ( ( a->x + a->w == b->x ) || ( b->x + b->w == a->x ) ) );
}

// destroys a & b
region *join( region *a, region *b )
{
	a->w = ( a->x == b->x ) ? a->w : a->w + b->w ;
	a->h = ( a->y == b->y ) ? a->h : a->h + b->h ;
	a->x = MIN( a->x, b->x );
	a->y = MIN( a->y, b->y );
	destroy_region( b );
	return a;
}

void grow( region **list, region *new )
{
	region *r = *list;

	while( r != NULL ){
		if( joinable( r, new ) ){
			remove_region( list, r );
			grow( list, join( r, new ) );
			return;
		}
		r = r->next;
	}
	add_region( list, new );
}




#define RISING   1
#define FALLING  (-1)
#define REGION_X 1
#define REGION_Y 0


void sort( region **list, int field, int order )
{
	region *sorted;
	region *max, *s;
	
	while( *list != NULL ){
		max = s = *list;
		while( s != NULL ){
			if( (field == REGION_X && (s->x - max->x)*order < 0)||
			    (field == REGION_Y && (s->y - max->y)*order < 0))
				max = s;
			s = s->next;
		}		
		remove_region( list, max );
		add_region( &sorted, max );
	}
	*list = sorted;
}

void order( region **list, int dx, int dy )
{
}



region *intersect( region *a, region *b )
{
	region *new = NULL;
	region *r;

	while( a != NULL ){
		r = b;
		while( r != NULL ){
			region *t = intersect_single( a, r );

			grow( &new, t );
			r = r->next;
		}
		a = a->next;
	}
	return new;
}

region *intersect_single(region *a, region *b)
{
	int x, y, w, h;

	x = MAX( a->x, b->x );
	y = MAX( a->y, b->y );


	if( a->x < b->x )
		w = MIN( a->x + a->w - b->x, b->w );
	else
		w = MIN( b->x + b->w - a->x, a->w );

	if( a->y < b->y )
		h = MIN( a->y + a->h - b->y, b->h );
	else
		h = MIN( b->y + b->h - a->y, a->h );

	if( w < 1 || h < 1 )
		return NULL;
	else
		return create_region( x, y, w, h );
}


void translate( region *list, int dx, int dy )
{
	while( list != NULL ){
		list->x += dx;
		list->y += dy;
		list = list->next;
	}
}

region *find_region( region *list, int x, int y, int w, int h )
{
	while( list != NULL ){
		if( list->x == x && list->y == y && 
		    list->w == w && list->h == h )
			return list;
		list = list->next;
	}
	return NULL;
}

region *duplicate( region *source )
{
	region *result = NULL;
	region *new;
	region *last = NULL;
	while( source != NULL ){
		new = create_region(source->x, source->y, 
				     source->w, source->h);
		if( last != NULL )
			last->next = new;
		else
			result = new;

		last = new;
		source = source->next;
	}
	return result;
}


void print_regions( region *list )
{
	int i = 0;
	region *r = list;
	if( r == NULL )
		printf("empty list\n");
		
	while( r != NULL ){
		printf("region %d: (%d %d) %d x %d    @ %p\n",
		       i++, r->x, r->y, r->w, r->h, r);
		r = r->next;
	}
}


region *remove_region( region **list, region *r )
{
	region *t = *list;
	if( r == NULL )
		return NULL;
	if( r == t ){
		*list =  r->next;
		r->next = NULL;
		return r;
	}
	while( t != NULL ){
		if( t->next == r){
			t->next = r->next;
			r->next = NULL;
			return r;
		}
		t = t->next;
	}
	return NULL;
}

void add_region( region **list, region *r )
{
	if( r == NULL )
		return;
	if( r->next != NULL ){
		printf("cannot add a list\n");
		exit(2);
	}
	r->next = *list;
	*list = r;
}

int n = 0;

void destroy_regions( region **r )
{
	region *t = *r;
	region *p;
	while( t != NULL ){
		p = t->next;
		remove_region(r, t);
		destroy_region( t );
		t = p;
	}
	*r = NULL;
}

int destroy_region( region *r )
{
	if( r == NULL )
		return 0;
	if( r->next != NULL ){
		printf("region locked in list \n");
		exit(0);
		return -1;
	}
	free( r );
	n--;
	return 0;
}

region *create_region( int x, int y, int w, int h )
{
/*	if( (w == 0) || (h == 0)){
		printf("ignoring zero region\n");
		return NULL;
	}
*/	region *new = (region *)malloc( sizeof( region ) );
	if( new != NULL ){
		new->x = x;
		new->y = y;
		new->w = w;
		new->h = h;
		new->next = NULL;
		n++;
	}
	return new;
}

